// Copyright 2019 Xanadu Quantum Technologies Inc.

// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at

//     http://www.apache.org/licenses/LICENSE-2.0

// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
/**
 * @file
 * Information regarding the libwalrus version number
 */
#pragma once

/// The major version number
#define HAFNIAN_VERSION_MAJOR 0

/// The minor version number
#define HAFNIAN_VERSION_MINOR 17

/// The patch number
#define HAFNIAN_VERSION_PATCH 0

/// The complete version number
#define HAFNIAN_VERSION_CODE (HAFNIAN_VERSION_MAJOR * 10000 + HAFNIAN_VERSION_MINOR * 100 + HAFNIAN_VERSION_PATCH)

/// Version number as string
#define HAFNIAN_VERSION_STRING "0.17.0"
